/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

#include "rules.h"
#include <mapidefs.h>
#include <mapicode.h>
#include <mapiutil.h>
#include <mapiext.h>
#include <edkmdb.h>
#include <edkguid.h>
#include "stringutil.h"
#include "Util.h"
#include "CommonUtil.h"
#include "mapi_ptr.h"
#include "str1252.h"

HRESULT MungeForwardBody(LPMESSAGE lpMessage)
{
	HRESULT hr = hrSuccess;
	SPropArrayPtr ptrBodies;
	SizedSPropTagArray (4, sBody) = { 4, {
			PR_BODY_A,
			PR_HTML,
			PR_RTF_IN_SYNC,
			PR_INTERNET_CPID
		} };
	SPropArrayPtr ptrInfo;
	SizedSPropTagArray (6, sInfo) = { 6, {
			// SENT_REPRESENTING?
			PR_SENDER_NAME_A,
			PR_SENDER_EMAIL_ADDRESS_A,
			PR_MESSAGE_DELIVERY_TIME,
			PR_RECEIVED_BY_NAME_A,
			PR_RECEIVED_BY_EMAIL_ADDRESS_A,
			PR_SUBJECT_A
		} };
	ULONG ulCharset;
	ULONG cValues;
	bool bPlain = false;
	SPropValue sNewBody;
	StreamPtr ptrStream;
	string strHTML;
	string strHTMLForwardText;
	string strBody;
	string strForwardText;

	hr = lpMessage->GetProps((LPSPropTagArray)&sBody, NULL, &cValues, &ptrBodies);
	if (FAILED(hr))
		goto exit;

	if (PROP_TYPE(ptrBodies[3].ulPropTag) != PT_ERROR)
		ulCharset = ptrBodies[3].Value.ul;
	else
		ulCharset = 20127;		// us-ascii

	if (PROP_TYPE(ptrBodies[0].ulPropTag) == PT_ERROR && PROP_TYPE(ptrBodies[1].ulPropTag) == PT_ERROR) {
		// plain and html not found, check sync flag
		bPlain = (ptrBodies[2].Value.b == FALSE);
	} else {
		bPlain = PROP_TYPE(ptrBodies[1].ulPropTag) == PT_ERROR && ptrBodies[1].Value.err == MAPI_E_NOT_FOUND;
	}
	sNewBody.ulPropTag = bPlain ? PR_BODY_A : PR_HTML;

	// From: <fullname>
	// Sent: <date>
	// To: <fullname>
	// Subject: <>
	// Auto forwarded by a rule

	hr = lpMessage->GetProps((LPSPropTagArray)&sInfo, NULL, &cValues, &ptrInfo);
	if (FAILED(hr))
		goto exit;

	if (bPlain) {
		// Plain text body

		strForwardText = "From: ";
		if (PROP_TYPE(ptrInfo[0].ulPropTag) != PT_ERROR)
			strForwardText += ptrInfo[0].Value.lpszA;
		else if (PROP_TYPE(ptrInfo[1].ulPropTag) != PT_ERROR)
			strForwardText += ptrInfo[1].Value.lpszA;

		strForwardText += "\nSent: ";
		if (PROP_TYPE(ptrInfo[2].ulPropTag) != PT_ERROR) {
			char buffer[64];
			time_t t;
			struct tm date;
			FileTimeToUnixTime(ptrInfo[2].Value.ft, &t);
			localtime_r(&t, &date);
			strftime(buffer, arraySize(buffer), "%c", &date);
			strForwardText += buffer;
		}

		strForwardText += "\nTo: ";
		if (PROP_TYPE(ptrInfo[3].ulPropTag) != PT_ERROR)
			strForwardText += ptrInfo[3].Value.lpszA;
		else if (PROP_TYPE(ptrInfo[4].ulPropTag) != PT_ERROR)
			strForwardText += ptrInfo[4].Value.lpszA;

		strForwardText += "\nSubject: ";
		if (PROP_TYPE(ptrInfo[5].ulPropTag) != PT_ERROR)
			strForwardText += ptrInfo[5].Value.lpszA;

		strForwardText += "\nAuto forwarded by a rule\n\n";

		if (ptrBodies[0].ulPropTag == PT_ERROR) {
			hr = lpMessage->OpenProperty(PR_BODY_A, &IID_IStream, 0, 0, &ptrStream);
			if (hr == hrSuccess) {
				hr = Util::HrStreamToString(ptrStream, strBody);
			}
			// stream
			strForwardText.append(strBody);
		} else {
			strForwardText += ptrBodies[0].Value.lpszA;
		}

		sNewBody.Value.lpszA = (char*)strForwardText.c_str();
	} else {
		// HTML body (or rtf, but nuts to editing that!)
		string strFind("<body");
		const char* pos;

		hr = lpMessage->OpenProperty(PR_HTML, &IID_IStream, 0, 0, &ptrStream);
		if (hr == hrSuccess) {
			hr = Util::HrStreamToString(ptrStream, strHTML);
		}
		// icase <body> tag 
		pos = strcasestr_str1252((char*)strHTML.c_str(), (char*)strFind.c_str());
		pos = pos ? pos + strFind.length() : strHTML.c_str();
		// if body tag was not found, this will make it be placed after the first tag, probably <html>
		while (*pos && *pos != '>') pos++;
		if (*pos == '>') pos++;

		// @todo Util::HrTextToHtml (char*, string &, int charset); (convert windows1252 -> ulcharset)
		{
			strHTMLForwardText = "<b>From:</b> ";
			if (PROP_TYPE(ptrInfo[0].ulPropTag) != PT_ERROR)
				Util::HrTextToHtml(ptrInfo[0].Value.lpszA, strHTMLForwardText, ulCharset);
			else if (PROP_TYPE(ptrInfo[1].ulPropTag) != PT_ERROR)
				Util::HrTextToHtml(ptrInfo[1].Value.lpszA, strHTMLForwardText, ulCharset);

			strHTMLForwardText += "<br><b>Sent:</b> ";
			if (PROP_TYPE(ptrInfo[2].ulPropTag) != PT_ERROR) {
				char buffer[32];
				time_t t;
				struct tm date;
				FileTimeToUnixTime(ptrInfo[2].Value.ft, &t);
				localtime_r(&t, &date);
				strftime(buffer, 32, "%c", &date);
				strHTMLForwardText += buffer;
			}

			strHTMLForwardText += "<br><b>To:</b> ";
			if (PROP_TYPE(ptrInfo[3].ulPropTag) != PT_ERROR)
				Util::HrTextToHtml(ptrInfo[3].Value.lpszA, strHTMLForwardText, ulCharset);
			else if (PROP_TYPE(ptrInfo[4].ulPropTag) != PT_ERROR)
				Util::HrTextToHtml(ptrInfo[4].Value.lpszA, strHTMLForwardText, ulCharset);

			strHTMLForwardText += "<br><b>Subject:</b> ";
			if (PROP_TYPE(ptrInfo[5].ulPropTag) != PT_ERROR)
				Util::HrTextToHtml(ptrInfo[5].Value.lpszA, strHTMLForwardText, ulCharset);

			strHTMLForwardText += "<br><b>Auto forwarded by a rule</b><br><hr><br>";
		}

		strHTML.insert((pos - strHTML.c_str()), strHTMLForwardText);

		sNewBody.Value.bin.cb = strHTML.size();
		sNewBody.Value.bin.lpb = (BYTE*)strHTML.c_str();
	}

	// set new body with forward markers
	hr = lpMessage->SetProps(1, &sNewBody, NULL);

exit:
	return hr;
}

/** 
 * Checks the rule recipient list for a possible loop, and filters
 * that recipient. Returns an error when no recipients are left after
 * the filter.
 * 
 * @param[in] lpMessage The original delivered message performing the rule action
 * @param[in] lpRuleRecipients The recipient list from the rule
 * @param[out] lppNewRecipients The actual recipient list to perform the action on
 * 
 * @return MAPI error code
 */
HRESULT CheckRecipients(ECLogger *lpLogger, LPADRBOOK lpAdrBook, IMAPIProp *lpMessage, LPADRLIST lpRuleRecipients, LPADRLIST *lppNewRecipients)
{
	HRESULT hr = hrSuccess;
	LPADRLIST lpRecipients = NULL;
	std::string strFromName, strFromType, strFromAddress;
	std::string strRuleName, strRuleType, strRuleAddress;

	hr = HrGetAddress(lpAdrBook, (IMessage*)lpMessage,
					  PR_SENT_REPRESENTING_ENTRYID, PR_SENT_REPRESENTING_NAME,
					  PR_SENT_REPRESENTING_ADDRTYPE, PR_SENT_REPRESENTING_EMAIL_ADDRESS,
					  strFromName, strFromType, strFromAddress);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get from address 0x%08X", hr);
		goto exit;
	}

	hr = MAPIAllocateBuffer(CbNewADRLIST(lpRuleRecipients->cEntries), (void**)&lpRecipients);
	if (hr != hrSuccess)
		goto exit;

	lpRecipients->cEntries = 0;

	for (ULONG i = 0; i < lpRuleRecipients->cEntries; i++) {
		hr = HrGetAddress(lpAdrBook, lpRuleRecipients->aEntries[i].rgPropVals, lpRuleRecipients->aEntries[i].cValues, PR_ENTRYID,
						  CHANGE_PROP_TYPE(PR_DISPLAY_NAME, PT_UNSPECIFIED), CHANGE_PROP_TYPE(PR_ADDRTYPE, PT_UNSPECIFIED), CHANGE_PROP_TYPE(PR_SMTP_ADDRESS, PT_UNSPECIFIED), 
						  strRuleName, strRuleType, strRuleAddress);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to get rule address 0x%08X", hr);
			goto exit;
		}

		if (strFromAddress == strRuleAddress) {
			lpLogger->Log(EC_LOGLEVEL_INFO, "Same user found in From and rule, blocking for loop protection");
			continue;
		}

		// copy recipient
		hr = Util::HrCopyPropertyArray(lpRuleRecipients->aEntries[i].rgPropVals, lpRuleRecipients->aEntries[i].cValues,
									   &lpRecipients->aEntries[lpRecipients->cEntries].rgPropVals, &lpRecipients->aEntries[lpRecipients->cEntries].cValues, true);
		if (hr != hrSuccess)
			goto exit;

		lpRecipients->cEntries++;
	}

	if (lpRecipients->cEntries == 0) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Loop protection blocked all recipients, skipping rule");
		hr = MAPI_E_UNABLE_TO_COMPLETE;
		goto exit;
	}

	if (lpRecipients->cEntries != lpRuleRecipients->cEntries) {
		lpLogger->Log(EC_LOGLEVEL_INFO, "Loop protection blocked some recipients");
	}

	*lppNewRecipients = lpRecipients;
	lpRecipients = NULL;

exit:
	if (lpRecipients)
		FreeProws((LPSRowSet)lpRecipients);

	return hr;
}

HRESULT CreateForwardCopy(ECLogger *lpLogger, LPADRBOOK lpAdrBook, LPMDB lpOrigStore, IMAPIProp *lpOrigMessage, LPADRLIST lpRuleRecipients, bool bDoPreserveSender, bool bDoNotMunge, LPMESSAGE *lppMessage)
{
	HRESULT hr = hrSuccess;
	LPMESSAGE lpFwdMsg = NULL;

	LPMAPIFOLDER lpOutbox = NULL;
	LPSPropValue lpOutboxEntryID = NULL;
	LPSPropValue lpSentMailEntryID = NULL;
	LPSPropTagArray lpExclude = NULL; // non-free
	ULONG ulObjType;
	LPADRLIST lpRecipients = NULL;

	SizedSPropTagArray (7, sExcludeFromCopyForward) = { 7, {
		PR_MESSAGE_RECIPIENTS,
		PR_TRANSPORT_MESSAGE_HEADERS,
		PR_SENT_REPRESENTING_ENTRYID,
		PR_SENT_REPRESENTING_NAME,
		PR_SENT_REPRESENTING_ADDRTYPE,
		PR_SENT_REPRESENTING_EMAIL_ADDRESS,
		PR_SENT_REPRESENTING_SEARCH_KEY,
	} };

	SizedSPropTagArray (2, sExcludeFromCopyRedirect) = { 2, {
		PR_MESSAGE_RECIPIENTS,
		PR_TRANSPORT_MESSAGE_HEADERS,
	} };

	LPSPropValue lpOrigSubject = NULL;
	SPropValue sForwardProps[4];
	string strSubject;

	if (lpRuleRecipients == NULL || lpRuleRecipients->cEntries == 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = CheckRecipients(lpLogger, lpAdrBook, lpOrigMessage, lpRuleRecipients, &lpRecipients);
	if (hr == MAPI_E_UNABLE_TO_COMPLETE)
		goto exit;
	if (hr != hrSuccess)
		// use rule recipients without filter
		lpRecipients = lpRuleRecipients;

	hr = HrGetOneProp(lpOrigStore, PR_IPM_OUTBOX_ENTRYID, &lpOutboxEntryID);
	if (hr != hrSuccess)
		goto exit;

	hr = HrGetOneProp(lpOrigStore, PR_IPM_SENTMAIL_ENTRYID, &lpSentMailEntryID);
	if (hr != hrSuccess)
		goto exit;

	hr = lpOrigStore->OpenEntry(lpOutboxEntryID->Value.bin.cb, (LPENTRYID)lpOutboxEntryID->Value.bin.lpb, NULL, MAPI_MODIFY, &ulObjType, (LPUNKNOWN*)&lpOutbox);
	if (hr != hrSuccess)
		goto exit;

	hr = lpOutbox->CreateMessage(NULL, 0, &lpFwdMsg);
	if (hr != hrSuccess)
		goto exit;

	// If we're doing a redirect, copy over the original PR_SENT_REPRESENTING_*, otherwise don't
	if(bDoPreserveSender)
		lpExclude = (LPSPropTagArray)&sExcludeFromCopyRedirect;
	else
		lpExclude = (LPSPropTagArray)&sExcludeFromCopyForward;
				
	hr = lpOrigMessage->CopyTo(0, NULL, lpExclude, 0, NULL, &IID_IMessage, lpFwdMsg, 0, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpFwdMsg->ModifyRecipients(MODRECIP_ADD, lpRecipients);
	if (hr != hrSuccess)
		goto exit;

	// set from email ??

	hr = HrGetOneProp(lpOrigMessage, PR_SUBJECT, &lpOrigSubject);
	if (hr == hrSuccess)
		strSubject = lpOrigSubject->Value.lpszA;

	if(!bDoNotMunge)
		strSubject = "FW: "+strSubject;

	sForwardProps[0].ulPropTag = PR_AUTO_FORWARDED;
	sForwardProps[0].Value.b = TRUE;

	sForwardProps[1].ulPropTag = PR_SUBJECT;
	sForwardProps[1].Value.lpszA = (char*)strSubject.c_str();

	sForwardProps[2].ulPropTag = PR_SENTMAIL_ENTRYID;
	sForwardProps[2].Value.bin.cb = lpSentMailEntryID->Value.bin.cb;
	sForwardProps[2].Value.bin.lpb = lpSentMailEntryID->Value.bin.lpb;

	hr = lpFwdMsg->SetProps(3, (LPSPropValue)&sForwardProps, NULL);
	if (hr != hrSuccess)
		goto exit;

	if (!bDoNotMunge)
		MungeForwardBody(lpFwdMsg);

	*lppMessage = lpFwdMsg;

exit:
	if (lpSentMailEntryID)
		MAPIFreeBuffer(lpSentMailEntryID);

	if (lpOrigSubject)
		MAPIFreeBuffer(lpOrigSubject);

	if (lpOutboxEntryID)
		MAPIFreeBuffer(lpOutboxEntryID);

	if (lpOutbox)
		lpOutbox->Release();

	if (lpRecipients && lpRecipients != lpRuleRecipients)
		FreeProws((LPSRowSet)lpRecipients);

	return hr;
}

// HRESULT HrDelegateMessage(LPMAPISESSION lpSession, LPEXCHANGEMANAGESTORE lpIEMS, IMAPIProp *lpMessage, LPADRENTRY lpAddress, ECLogger *lpLogger)
HRESULT HrDelegateMessage(IMAPIProp *lpMessage)
{
	HRESULT hr = hrSuccess;
	SPropValue sNewProps[6] = {{0}};
	LPSPropValue lpProps = NULL;
	ULONG cValues = 0;
	SizedSPropTagArray(5, sptaRecipProps) = {5, 
		{ PR_RECEIVED_BY_ENTRYID, PR_RECEIVED_BY_ADDRTYPE, PR_RECEIVED_BY_EMAIL_ADDRESS, PR_RECEIVED_BY_NAME, PR_RECEIVED_BY_SEARCH_KEY }
	};
	SizedSPropTagArray(1, sptaSentMail) = { 1, { PR_SENTMAIL_ENTRYID } };

	// set PR_RCVD_REPRESENTING on original receiver
	hr = lpMessage->GetProps((LPSPropTagArray)&sptaRecipProps, 0, &cValues, &lpProps);
	if (hr != hrSuccess)
		goto exit;

	lpProps[0].ulPropTag = PR_RCVD_REPRESENTING_ENTRYID;
	lpProps[1].ulPropTag = PR_RCVD_REPRESENTING_ADDRTYPE;
	lpProps[2].ulPropTag = PR_RCVD_REPRESENTING_EMAIL_ADDRESS;
	lpProps[3].ulPropTag = PR_RCVD_REPRESENTING_NAME;
	lpProps[4].ulPropTag = PR_RCVD_REPRESENTING_SEARCH_KEY;

	hr = lpMessage->SetProps(cValues, lpProps, NULL);
	if (hr != hrSuccess)
		goto exit;

	// TODO: delete PR_RECEIVED_BY_ values?

	sNewProps[0].ulPropTag = PR_DELEGATED_BY_RULE;
	sNewProps[0].Value.b = TRUE;

	sNewProps[1].ulPropTag = PR_DELETE_AFTER_SUBMIT;
	sNewProps[1].Value.b = TRUE;

	hr = lpMessage->SetProps(2, (LPSPropValue)&sNewProps, NULL);
	if (hr != hrSuccess)
		goto exit;
		
	// Don't want to move to sent mail
	hr = lpMessage->DeleteProps((LPSPropTagArray)&sptaSentMail, NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE);

exit:
	if (lpProps)
		MAPIFreeBuffer(lpProps);

	return hr;
}


// lpMessage: gets EntryID, maybe pass this and close message in DAgent.cpp
HRESULT HrProcessRules(LPMAPISESSION lpSession, LPADRBOOK lpAdrBook, LPMDB lpOrigStore, LPMAPIFOLDER lpOrigInbox, IMessage **lppMessage, ECLogger *lpLogger) {
	HRESULT hr = hrSuccess;
    IExchangeModifyTable *lpTable = NULL;
    IMAPITable *lpView = NULL;

	LPMDB lpDestStore = NULL;
	LPMAPIFOLDER lpDestFolder = NULL;
	LPMESSAGE lpReplyMsg = NULL;
	LPMESSAGE lpFwdMsg = NULL;
	ULONG ulObjType;
	bool bAddFwdFlag = false;
	IMessage *lpNewMessage = NULL;

    LPSRowSet lpRowSet = NULL;
    SizedSPropTagArray(11, sptaRules) = {11,
										 { PR_RULE_ID, PR_RULE_IDS, PR_RULE_SEQUENCE, PR_RULE_STATE,
										   PR_RULE_USER_FLAGS, PR_RULE_CONDITION, PR_RULE_ACTIONS,
										   PR_RULE_PROVIDER, PR_RULE_NAME, PR_RULE_LEVEL, PR_RULE_PROVIDER_DATA } };
	SizedSSortOrderSet(1, sosRules) = {1, 0, 0, { {PR_RULE_SEQUENCE, TABLE_SORT_ASCEND} } };
    LPSPropValue lpRuleName = NULL;
    LPSPropValue lpRuleState = NULL;
	std::string strRule;
    LPSPropValue lpProp = NULL;
    LPSRestriction lpCondition = NULL;
    ACTIONS* lpActions = NULL;

	SPropValue sForwardProps[4];


    hr = lpOrigInbox->OpenProperty(PR_RULES_TABLE, &IID_IExchangeModifyTable, 0, 0, (LPUNKNOWN *)&lpTable);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpTable->GetTable(0, &lpView);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpView->SetColumns((LPSPropTagArray)&sptaRules, 0);
    if (hr != hrSuccess)
        goto exit;

	hr = lpView->SortTable((LPSSortOrderSet)&sosRules, 0);
    if (hr != hrSuccess)
        goto exit;

	while (TRUE) {

        hr = lpView->QueryRows(1, 0, &lpRowSet);
		if (hr != hrSuccess)
			goto exit;

        if (lpRowSet->cRows == 0)
            break;

        lpRuleName = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_RULE_NAME);
		if (lpRuleName)
			strRule = lpRuleName->Value.lpszA;
		else
			strRule = "(no name)";

		lpRuleState = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_RULE_STATE);
		if (lpRuleState) {
			if (!(lpRuleState->Value.i & ST_ENABLED)) {
				lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule '%s' is disabled, skipping...", strRule.c_str());
				goto nextrule;		// rule is disabled
			}
		}

		lpCondition = NULL;
		lpActions = NULL;

		lpProp = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_RULE_CONDITION);
		if (lpProp)
			// NOTE: object is placed in Value.lpszA, not Value.x
			lpCondition = (LPSRestriction)lpProp->Value.lpszA;
		if (!lpCondition) {
			lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule '%s' has no contition, skipping...", strRule.c_str());
			goto nextrule;
		}

		lpProp = PpropFindProp(lpRowSet->aRow[0].lpProps, lpRowSet->aRow[0].cValues, PR_RULE_ACTIONS);
		if (lpProp)
			// NOTE: object is placed in Value.lpszA, not Value.x
			lpActions = (ACTIONS*)lpProp->Value.lpszA;
		if (!lpActions) {
			lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule '%s' has no action, skipping...", strRule.c_str());
			goto nextrule;
		}
		
		// test if action should be done...
		if (TestRestriction(lpCondition, *lppMessage) != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_INFO, (std::string)"Rule " + strRule + " doesn't match");
			goto nextrule;
		}	
		lpLogger->Log(EC_LOGLEVEL_INFO, (std::string)"Rule " + strRule + " matches");

		for(ULONG n=0; n<lpActions->cActions; n++) {

			// do action
			switch(lpActions->lpAction[n].acttype) {
			case OP_MOVE:
			case OP_COPY:
				if (lpActions->lpAction[n].acttype == OP_COPY)
					lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: copying e-mail");
				else
					lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: moving e-mail");

				// First try to open the folder on the session as that will just work if we have the store open
				hr = lpSession->OpenEntry(lpActions->lpAction[n].actMoveCopy.cbFldEntryId,
										  lpActions->lpAction[n].actMoveCopy.lpFldEntryId, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType,
										  (IUnknown**)&lpDestFolder);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_INFO, (std::string)"Rule "+strRule+": Unable to open folder through session, trying through store");

					hr = lpSession->OpenMsgStore(0, lpActions->lpAction[n].actMoveCopy.cbStoreEntryId,
													lpActions->lpAction[n].actMoveCopy.lpStoreEntryId, NULL, MAPI_BEST_ACCESS, &lpDestStore);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": Unable to open destination store");
						goto nextact;
					}

					hr = lpDestStore->OpenEntry(lpActions->lpAction[n].actMoveCopy.cbFldEntryId,
												lpActions->lpAction[n].actMoveCopy.lpFldEntryId, &IID_IMAPIFolder, MAPI_MODIFY, &ulObjType,
												(IUnknown**)&lpDestFolder);
					if (hr != hrSuccess || ulObjType != MAPI_FOLDER) {
						lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": Unable to open destination folder");
						goto nextact;
					}
				}

				hr = lpDestFolder->CreateMessage(NULL, 0, &lpNewMessage);
				if(hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to create e-mail for rule " + strRule);
					goto exit;
				}
					
				hr = (*lppMessage)->CopyTo(0, NULL, NULL, 0, NULL, &IID_IMessage, lpNewMessage, 0, NULL);
				if(hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to copy e-mail for rule " + strRule);
					goto exit;
				}
				// @todo: PR_EC_IMAP_EMAIL single instance

				if (lpActions->lpAction[n].acttype == OP_MOVE) {
					// Release the original message, it will disappear unsaved
					(*lppMessage)->Release();
					*lppMessage = lpNewMessage;
					lpNewMessage = NULL;
				} else {
					// Save the copy in its new location
					hr = lpNewMessage->SaveChanges(0);
				}
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": Unable to copy/move message");
					goto nextact;
				}

				break;

			// may become dam's, may become normal rules (ol2k3)

			case OP_REPLY:
			case OP_OOF_REPLY:
// FIX TODO FIRST
#if 0
				if (lpActions->lpAction[n].acttype == OP_REPLY)
					lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: replying e-mail");
				else
					lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: OOF replying e-mail");

				hr = lpOrigInbox->OpenEntry(lpActions->lpAction[n].actReply.cbEntryId,
											lpActions->lpAction[n].actReply.lpEntryId, &IID_IMessage, 0, &ulObjType,
											(IUnknown**)&lpReplyMsg);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": Unable to open reply message");
					goto nextact;
				}
				// TODO:
				// 1. make copy of lpReplyMsg message, needs CopyTo() function
				// 2. copy From: to To:
				hr = lpReplyMsg->SubmitMessage(0);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": Unable to send reply message");
					goto nextact;
				}
#endif				
				lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": REPLY actions are currently unsupported");
				break;

			case OP_FORWARD:
				// TODO: test lpActions->lpAction[n].ulActionFlavor
				// FWD_PRESERVE_SENDER			1
				// FWD_DO_NOT_MUNGE_MSG			2
				// FWD_AS_ATTACHMENT			4

				// redirect == 3

				if (lpActions->lpAction[n].lpadrlist->cEntries == 0) {
					lpLogger->Log(EC_LOGLEVEL_DEBUG, "Forwarding rule doesn't have recipients");
					continue; // Nothing todo
				}

				lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: forwarding e-mail");

				hr = CreateForwardCopy(lpLogger, lpAdrBook, lpOrigStore, *lppMessage, lpActions->lpAction[n].lpadrlist,
									   lpActions->lpAction[n].ulActionFlavor & FWD_PRESERVE_SENDER,
									   lpActions->lpAction[n].ulActionFlavor & FWD_DO_NOT_MUNGE_MSG, &lpFwdMsg);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": FORWARD Unable to create forward message");
					goto nextact;
				}

				hr = lpFwdMsg->SubmitMessage(0);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": FORWARD Unable to send forward message");
					goto nextact;
				}

				// update original message, set as forwarded
				bAddFwdFlag = true;
				break;

			case OP_BOUNCE:
				// scBounceCode?
				// TODO:
				// 1. make copy of lpMessage, needs CopyTo() function
				// 2. copy From: to To:
				// 3. SubmitMessage()
				lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": BOUNCE actions are currently unsupported");
				break;

			case OP_DELEGATE:
				
				if (lpActions->lpAction[n].lpadrlist->cEntries == 0) {
					lpLogger->Log(EC_LOGLEVEL_DEBUG, "Delegating rule doesn't have recipients");
					continue; // Nothing todo
				}
				lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: delegating e-mail");

				hr = CreateForwardCopy(lpLogger, lpAdrBook, lpOrigStore, *lppMessage, lpActions->lpAction[n].lpadrlist, true, true, &lpFwdMsg);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": DELEGATE Unable to create delegate message");
					goto nextact;
				}

				// set delegate properties
				hr = HrDelegateMessage(lpFwdMsg);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": DELEGATE Unable to modify delegate message");
					goto nextact;
				}

				hr = lpFwdMsg->SubmitMessage(0);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": DELEGATE Unable to send delegate message");
					goto nextact;
				}

				// don't set forwarded flag
				break;

			// will become a DAM atm, so I won't even bother implementing these ...

			case OP_DEFER_ACTION:
				// DAM crud, but outlook doesn't check these messages .. yet
				lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": DEFER client actions are currently unsupported");
				break;
			case OP_TAG:
				// sure .. WHEN YOU STOP WITH THE FRIGGIN' DEFER ACTION MESSAGES!! dork!
				lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": TAG actions are currently unsupported");
				break;
			case OP_DELETE:
				// since *lppMessage wasn't yet saved in the server, we can just return a special MAPI Error code here,
				// this will trigger the out-of-office mail (according to microsoft), but not save the message and drop it.
				// The error code will become hrSuccess automatically after returning from the post processing function.
				lpLogger->Log(EC_LOGLEVEL_DEBUG, "Rule action: deleting e-mail");
				hr = MAPI_E_CANCEL;
				goto exit;
				break;
			case OP_MARK_AS_READ:
				// add prop read
				lpLogger->Log(EC_LOGLEVEL_FATAL, (std::string)"Rule "+strRule+": MARK AS READ actions are currently unsupported");
				break;
			};

			// next action
nextact:
			if (lpDestStore) {
				lpDestStore->Release();
				lpDestStore = NULL;
			}
			if (lpDestFolder) {
				lpDestFolder->Release();
				lpDestFolder = NULL;
			}
			if (lpReplyMsg) {
				lpReplyMsg->Release();
				lpReplyMsg = NULL;
			}
			if (lpFwdMsg) {
				lpFwdMsg->Release();
				lpFwdMsg = NULL;
			}
			if (lpNewMessage) {
				lpNewMessage->Release();
				lpNewMessage = NULL;
			}
		} // end action loop

		if (lpRuleState && (lpRuleState->Value.i & ST_EXIT_LEVEL))
			break;

nextrule:
		if (lpRowSet) {
			FreeProws(lpRowSet);
			lpRowSet = NULL;
		}
	}

	if (bAddFwdFlag) {
		sForwardProps[0].ulPropTag = PR_ICON_INDEX;
		sForwardProps[0].Value.ul = ICON_MAIL_FORWARDED;

		sForwardProps[1].ulPropTag = PR_LAST_VERB_EXECUTED;
		sForwardProps[1].Value.ul = NOTEIVERB_FORWARD;

		sForwardProps[2].ulPropTag = PR_LAST_VERB_EXECUTION_TIME;
		GetSystemTimeAsFileTime(&sForwardProps[2].Value.ft);

		// set forward in msg flag
		hr = (*lppMessage)->SetProps(3, (LPSPropValue)&sForwardProps, NULL);
	}

exit:
	if (hr != hrSuccess && hr != MAPI_E_CANCEL)
		lpLogger->Log(EC_LOGLEVEL_INFO, "Error while processing rules: 0x%08X", hr);

	if (lpNewMessage)
		lpNewMessage->Release();
		
	if (lpTable)
		lpTable->Release();

	if (lpView)
		lpView->Release();

	if (lpRowSet)
		FreeProws(lpRowSet);

	return hr;
}
